from typing import Any, List, Tuple
from wsgiref.simple_server import make_server

_URLPATTERNS: List[Tuple[str, Any]] = list()


def not_found(environ, start_response):
    """Called if no URL matches."""
    start_response("404 NOT FOUND", [("Content-Type", "text/plain")])
    return ["Not Found".encode()]


def urlpatterns(patterns: List[Tuple[str, Any]]):
    global _URLPATTERNS
    _URLPATTERNS = patterns


def get_patterns():
    return _URLPATTERNS


def serve(host: str, ip: str, port: int, application: Any, debug: bool = True):
    if debug:
        _host: str = ip if host is None else host

        with make_server(_host, port, application) as server:
            print("Server listening on port http://localhost:8000...")
            try:
                server.serve_forever()
            except KeyboardInterrupt:
                server.shutdown()
                return
    pass



import os
import sys
from importlib import import_module
from inspect import getmembers, isfunction
from pathlib import Path
from types import ModuleType

from typing import List

from aquilo.main import Aquilo
from aquilo.utils import has_special_char

settings = import_module(os.environ.get("AQUILO_SETTINGS_MODULE"))

app = Aquilo(debug=settings.DEBUG)


def validate_home_page(application: str, apps_list: List[str], module: ModuleType):
    if application == apps_list[0]:
        homepage_function = list()

        for member in getmembers(module, isfunction):
            if member[0].startswith("page_"):
                name = member[1].__name__.split("page_")

                if len(name) == 1:
                    raise ValueError("Invalid page name.")
                homepage_function.append(member[1])

        if len(homepage_function) > 1:
            raise ValueError(
                "The homepage application must have only one function."
                "Please remove the other functions."
            )


def runserver_command():
    apps_list: List[str] = settings.APPS

    if len(apps_list) == 0:
        raise ValueError("No apps specified." "Please specify apps in settings.APPS.")

    for application in apps_list:
        try:
            app_module = import_module(
                f"{os.environ.get('AQUILO_APPS_MODULE')}.{application}.pages"
            )
        except ImportError as exc:
            raise ValueError(
                f"{application} module was not found in apps. "
                "Please check the spelling."
            ) from exc
        validate_home_page(application, apps_list, app_module)
        register_pages(application, app_module)

    app.run()


def register_pages(application: str, module: ModuleType):
    for member in getmembers(module, isfunction):
        if member[0].startswith("page_"):
            name = member[1].__name__.split("page_")

            if len(name) == 1:
                raise ValueError("Invalid page name.")

            member[1].__name__ = f"page_{application}_{name[1]}"
            app.page(member[1])


def startapp_command(app_name: str):
    if " " in app_name:
        raise ValueError("App name cannot contain spaces.")

    if app_name == "aquilo":
        raise ValueError("Aquilo is a reserved name.")

    if app_name in settings.APPS:
        raise ValueError("App already exists.")

    if has_special_char(app_name):
        raise ValueError("App name cannot contain special characters.")

    app_directory = os.environ.get("AQUILO_APPS_MODULE").replace(".", os.sep)
    new_app_directory = Path(app_directory + os.sep + app_name)

    if os.path.exists(new_app_directory):
        raise ValueError("Directory already exists.")

    os.makedirs(new_app_directory)

    with open(f"{new_app_directory}/__init__.py", "w") as init:
        init.write("")

    with open(f"{new_app_directory}/pages.py", "w") as pages:
        pages.write(f"""from aquilo import build\n\n
def page_{app_name}():
    pass\n""")


def format_code():
    os.system(f"black {settings.BASE_DIR}")


def execute_from_command_line(args: List[str]) -> None:
    if len(args) == 1:
        raise ValueError("No command specified.")

    command: str = args[1]
    if command == "runserver":
        runserver_command()
    elif command == "startapp":
        try:
            startapp_command(args[2])
        except IndexError:
            raise ValueError("No app name specified.")
    elif command == "format":
        format_code()
    else:
        raise ValueError("Invalid command.")


def main():
    execute_from_command_line(sys.argv)



from pathlib import Path

from aquilo.browser.elements.containers import *
from aquilo.browser.elements.typography import *
from aquilo.html import *
from aquilo.main import *
from aquilo.utils import has_special_char


def startproject(project_name: str):
    if has_special_char(project_name):
        raise ValueError("Project name cannot contain special characters.")

    if project_name == "aquilo":
        raise ValueError("Aquilo is a reserved name.")

    if " " in project_name:
        raise ValueError("Project name cannot contain spaces.")

    new_app_directory = Path(os.getcwd() + os.sep + project_name)

    if os.path.exists(new_app_directory):
        raise ValueError("Project %s already exists." % project_name)

    os.makedirs(new_app_directory)

    with open(f"{os.getcwd()}/manage.py", "w") as manage_file:
        manage_file.write(
            f"""#!/usr/bin/env python
\"\"\"Aquilo's command-line utility for administrative tasks.\"\"\"
import os
import sys


def main():
    \"\"\"Run administrative tasks.\"\"\"
    os.environ.setdefault(\"AQUILO_SETTINGS_MODULE\", \"{project_name}.config.settings\")
    os.environ.setdefault(\"AQUILO_APPS_MODULE\", \"{project_name}.apps\")

    try:
        from aquilo.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            \"Couldn't import Aquilo. Are you sure it's installed and \"
            \"available on your PYTHONPATH environment variable? Did you \"
            \"forget to activate a virtual environment?\"
        ) from exc
    execute_from_command_line(sys.argv)


if __name__ == \"__main__\":
    main()
"""
        )

    with open(f"{os.getcwd()}/.gitignore", "w") as gitignore_file:
        gitignore_file.write(
            """# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# C extensions
*.so

# Distribution / packaging
.Python
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST
build/

# PyInstaller
#  Usually these files are written by a python script from a template
#  before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec

# Installer logs
pip-log.txt
pip-delete-this-directory.txt

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/
cover/

# Translations
*.mo
*.pot

# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal

# Flask stuff:
instance/
.webassets-cache

# Scrapy stuff:
.scrapy

# Sphinx documentation
docs/_build/

# PyBuilder
.pybuilder/
target/

# Jupyter Notebook
.ipynb_checkpoints

# IPython
profile_default/
ipython_config.py

# pyenv
#   For a library or package, you might want to ignore these files since the code is
#   intended to run in multiple environments; otherwise, check them in:
# .python-version

# pipenv
#   According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.
#   However, in case of collaboration, if having platform-specific dependencies or dependencies
#   having no cross-platform support, pipenv may install dependencies that don't work, or not
#   install all needed dependencies.
#Pipfile.lock

# poetry
#   Similar to Pipfile.lock, it is generally recommended to include poetry.lock in version control.
#   This is especially recommended for binary packages to ensure reproducibility, and is more
#   commonly ignored for libraries.
#   https://python-poetry.org/docs/basic-usage/#commit-your-poetrylock-file-to-version-control
#poetry.lock

# pdm
#   Similar to Pipfile.lock, it is generally recommended to include pdm.lock in version control.
#pdm.lock
#   pdm stores project-wide configurations in .pdm.toml, but it is recommended to not include it
#   in version control.
#   https://pdm.fming.dev/#use-with-ide
.pdm.toml

# PEP 582; used by e.g. github.com/David-OConnor/pyflow and github.com/pdm-project/pdm
__pypackages__/

# Celery stuff
celerybeat-schedule
celerybeat.pid

# SageMath parsed files
*.sage.py

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/

# Spyder project settings
.spyderproject
.spyproject

# Rope project settings
.ropeproject

# mkdocs documentation
/site

# mypy
.mypy_cache/
.dmypy.json
dmypy.json

# Pyre type checker
.pyre/

# pytype static type analyzer
.pytype/

# Cython debug symbols
cython_debug/

# PyCharm
#  JetBrains specific template is maintained in a separate JetBrains.gitignore that can
#  be found at https://github.com/github/gitignore/blob/main/Global/JetBrains.gitignore
#  and can be added to the global gitignore or merged into this file.  For a more nuclear
#  option (not recommended) you can uncomment the following to ignore the entire idea folder.
#.idea/
        """
        )

    with open(f"{os.getcwd()}/README.md", "w") as readme_file:
        readme_file.write(
            f"""# {project_name.capitalize()}

This is a new Aquilo project.
        """
        )

    with open(f"{os.getcwd()}/requirements.txt", "w") as requirements_file:
        requirements_file.write("aquilo")

    with open(f"{os.getcwd()}/{project_name}/__init__.py", "w") as init_file:
        init_file.write("")

    os.makedirs(f"{os.getcwd()}/{project_name}/apps")

    with open(f"{os.getcwd()}/{project_name}/apps/__init__.py", "w") as init_file:
        init_file.write("")

    os.makedirs(f"{os.getcwd()}/{project_name}/apps/home")
    with open(f"{os.getcwd()}/{project_name}/apps/home/__init__.py", "w") as init_file:
        init_file.write("")

    with open(f"{os.getcwd()}/{project_name}/apps/home/pages.py", "w") as pages:
        pages.write("""from aquilo import build, div, h1


def page_home():
    root = div(h1(\"Hello World!\"))
    return build(root, title=\"Hello World!\")
""")

    os.makedirs(f"{os.getcwd()}/{project_name}/config")

    with open(f"{os.getcwd()}/{project_name}/config/__init__.py", "w") as init_file:
        init_file.write("")

    with open(f"{os.getcwd()}/{project_name}/config/settings.py", "w") as settings_file:
        settings_file.write(
            f"""from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent.parent

DEBUG = True

APPS = [
    "home",
]

ROOT_URLCONF = "{project_name}.config.urls"\n
"""
        )

    os.makedirs(f"{os.getcwd()}/{project_name}/libs")

    with open(f"{os.getcwd()}/{project_name}/libs/__init__.py", "w") as init_file:
        init_file.write("")

    try:
        os.system(f"git init")
    except:
        pass


def main():
    import sys

    if len(sys.argv) == 1:
        raise ValueError("No command specified.")

    command: str = sys.argv[1]
    if command == "startproject":
        try:
            startproject(sys.argv[2].lower())
        except IndexError:
            raise ValueError("No project name specified.")
    else:
        raise ValueError("Invalid command.")


if __name__ == "__main__":
    main()



from typing import Union, Iterable

from aquilo.browser.elements import Element
from aquilo.html.generators import generate_dom_tree


class div(Element):
    def __init__(
        self,
        *args: Union[Iterable, Element],
    ):
        etype = self.__class__.__name__
        super().__init__(etype)
        self.elements = args

    def get_elements(self):
        return self.elements

    def __call__(self, *args, **kwargs):
        return generate_dom_tree(self.elements)



from typing import Union, Iterable, Callable

from aquilo.browser.elements import Element


def build(element_tree: Callable, title: str = None, description: str = None) -> str:
    """
    It takes an element tree and returns an HTML document

    :param element_tree: str
    :type element_tree: str
    :param title: The title of the page
    :type title: str
    :param description: str = None
    :type description: str
    :return: A string of HTML code.
    """
    html: str = f"""
<!DOCTYPE html><html lang="en"><head>
    <meta charset="UTF-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{title}</title>
</head><body>{element_tree()}</body></html>
        """
    return html


def get_404():
    return """
<!doctype html>
<html lang="en">
<head>
  <title>Not Found</title>
</head>
<body>
  <h1>Not Found</h1><p>The requested resource was not found on this server.</p>
</body>
</html>
""".encode()


def generate_dom_tree(root: Union[Iterable, Element]) -> str:
    """
    It takes a root element, and returns a string representation of the DOM tree

    :param root: The root element of the DOM tree
    :type root: Union[Iterable, Element]
    :return: A string of HTML code.
    """
    element_tree = ["<div>"]

    for tag in root:
        try:
            iter(tag)
            is_iterable = True
        except TypeError:
            is_iterable = False

        if is_iterable:
            for i in tag:
                element_tree.append(i())
        else:
            element_tree.append(tag())

    element_tree.append("</div>")

    return "".join(element_tree)



from aquilo.browser.elements import Element


class Text(Element):
    def __init__(
        self,
        etype: str,
        text: str,
        eid: str = None,
    ):
        super().__init__(etype, text, eid)


class h1(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class h2(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class h3(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class h4(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class h5(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class h6(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class p(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class strong(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class small(Text):
    def __init__(
        self,
        text: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)


class a(Text):
    def __init__(
        self,
        text: str,
        href: str,
        eid: str = None,
    ):
        super().__init__(self.__class__.__name__, text, eid)
        self.href = href

    def __call__(self, *args, **kwargs):
        return f'<{self.etype} href="{self.href}">{self.innerHTML}</{self.etype}>'



import os
from typing import Any, Callable, List, Tuple, Dict

from aquilo.browser.elements.containers import div
from aquilo.html import get_404, build
from aquilo.http import serve, get_patterns, urlpatterns


class Aquilo:
    def __init__(
        self,
        host: str = None,
        ip: str = "127.0.0.1",
        port: int = 8000,
        debug: bool = True,
    ):
        self.host = host
        self.ip = ip
        self.port = port
        self.element_tree = None
        self._pages: Dict[str, Any] = {}
        self.root = None
        self._patterns: List[Tuple[str, Callable]] = list()
        self.debug: bool = debug

    def page(self, function: Callable) -> Callable:
        function_name_with_dashes: str = function.__name__.replace("_", "/").lower()
        name: List[str] = function_name_with_dashes.split("page/")

        if len(name) == 1:
            raise ValueError("Invalid page name.")

        path_name: str = "{}".format(name[1] + "/")

        try:
            path_list: List[str] = path_name.split("/")[:-1]

            if path_list[0] == path_list[1]:
                path_name = path_list[0] + "/"

        except IndexError:
            pass

        self._pages[path_name] = {"function": function}

        pattern: Tuple[str, Callable] = (path_name, function)
        self._patterns.append(pattern)

        return function

    def register_root(self, root: div) -> None:
        if not isinstance(root, div):
            raise TypeError("root must be a div")
        self.root: div = root

    def _application(self, environ, start_response) -> List[bytes]:
        start_response("200 OK", [("Content-Type", "text/html")])

        path: str = environ.get("PATH_INFO", "").lstrip("/")

        if not path.endswith("/"):
            path = path + "/"

        pages: List[str] = list(self._pages.keys())

        if path == "/" or path == "":
            html = self._pages[pages[0]]["function"]().encode()
            return [html]

        for pattern in get_patterns():
            if path in pages:
                html = self._pages[path]["function"]().encode()
                return [html]

        return [get_404()]

    def run(self) -> None:
        urlpatterns(self._patterns)
        serve(self.host, self.ip, self.port, self._application, self.debug)

    def __call__(self, *args, **kwargs):
        print(args, kwargs)
        return self._application(*args)

    def build_html(self, dir_path):
        if os.path.exists(os.path.join(dir_path, "build")):
            import shutil

            shutil.rmtree(os.path.join(dir_path, "build"), ignore_errors=True)

        os.mkdir(os.path.join(dir_path, "build"))

        pages = list(self._pages.keys())

        for page in pages:
            with open(os.path.join(dir_path, "build", f"{page}.html"), "w") as file:
                file.write(build(self._pages[page]["function"], title=page))



class Element:
    def __init__(self, etype: str, text: str = None, eid: str = None):
        self.etype = etype
        self.hidden: bool = False
        self.eid: str = eid
        self.innerHTML: str = text
        self.innerText: str = text
        self.tagName: str = etype.upper()
        self.title: str = ""

    def __call__(self, *args, **kwargs):
        return f"<{self.etype}>{self.innerHTML}</{self.etype}>"
